"""
JSON database class for chattimarkov.

JSON files are actually not really suitable for storing a key-value database.
Loading and storing can actually be slow, mostly if the chattimarkov instance
is learning a lot.
"""

import atexit
import json
import os.path

from .memory import MemoryDatabase


class JSONFileDatabase(MemoryDatabase):
    def __init__(self, filepath, *args, **kwargs):
        self.filepath = filepath
        if os.path.exists(self.filepath):
            db = json.load(open(filepath))
        else:
            db = None
        super().__init__(self, db, *args, **kwargs)
        atexit.register(self.cleanup)

    def cleanup(self):
        with open(self.filepath, "w") as stream:
            json.dump(self.db, stream)



"""Chattymarkov base code for database submodule.

Each time you want to add some support for a database, you must inherit this
class."""


class AbstractDatabase:
    """AbstractDatabase class."""

    def __init__(self, *args, **kwargs):
        pass

    def add(self, key, element):
        """Add an entry into the database."""

    def random(self, key):
        """Pick up a random entry from the `key` subset into the database."""

    def get(self, key, default=None):
        """Get the value associated to `key` into the database."""

    def set(self, key, value):
        """Set `value` to `key` in database."""



#!/usr/bin/env python3
from chattymarkov import ChattyMarkov
from chattymarkov.database import MemoryDatabase

db = MemoryDatabase()
markov = ChattyMarkov(db)
markov.learn("My favorite animal is the crocodile")
markov.learn("The word animal is six letters long")
print(markov.generate())



"""Memory database class for chattimarkov.

This is just a volatile, in-memory database which is built either from a
pre-existing dictionary or from scratch. Upon object destruction, the database
is not saved.
"""

import random

from .base import AbstractDatabase


class MemoryDatabase(AbstractDatabase):
    def __init__(self, db=None):
        if db is None:
            self.db = {}
        else:
            self.db = db

    def add(self, key, element):
        if key not in self.db:
            self.db[key] = []
        if type(self.db[key]) is not list:
            return False

        if element not in self.db[key]:
            self.db[key].append(element)

    def random(self, key):
        if key not in self.db:
            return None
        if type(self.db[key]) is not list:
            return None
        return random.choice(self.db[key])

    def get(self, key, default=None):
        return self.db.get(key, default)

    def set(self, key, value):
        self.db[key] = value



"""Chattymarkov database submodule.

This submodule gathers all the supported database formats.
"""

# flake8: noqa
from .json import JSONFileDatabase
from .memory import MemoryDatabase
from .redis import RedisDatabase



"""
ChattyMarkov module.

This module provide a base class to instanciate a `ChattyMarkov` class
which is responsible for learning from sentences as much as generating ones.

In order to succesfully memorise sentences, A `ChattyMarkov` instance will
rely on a database which has to inherit from the abstract interface described
`chattymarkov.database.AbstractDatabase`.
"""


class ChattyMarkov:
    """ChattyMarkov class.

    Define an interface between a user and a database in order to learn
    sentences and generate random sentences through a markov-chain-based
    algorithm."""

    def __init__(self, database, prefix="chattymarkov", separator="\x01",
                 stop_word="\x02"):
        """Instanciate the ChattyMarkov class.

        Args:
            - database: a database instance which inherits from
                        AbstractDatabase.
            - prefix: a prefix useful in database storage.
            - separator: a separator pattern for database storage.
            - stop_word: a stop-word pattern for database storage.
        """
        self.db = database
        self.separator = separator
        self.stop_word = stop_word
        self.prefix = prefix

    def _make_key(self, key):
        """Private method. Generate a key for internal database storage,
        given the `key` parameter.

        Args:
            - key: the string to generate the database key from.

        Returns: a key useful for internal use.
        """
        return '-'.join((self.prefix, key))

    def learn(self, msg):
        """Learn from a message. This function is called in order to
        memorise a sentence provided through the parameter `msg`.

        Args:
            - msg: the sentence to learn from.
        """
        if msg == '':
            return
        for words in self._split_message(msg):
            key = self.separator.join(words[:-1])
            self.db.add(self._make_key(key), words[-1])

    def generate(self):
        """Generate a message by browsing the database randomly as we
        browse a markov graph, to construct a random sentence from what
        the ChattyMarkov instance has learned so far.

        Returns: a string which consist of a random generated sentence.
        """
        lastword = ""
        previous = ""
        out = []

        while True:
            key = self._make_key(self.separator.join([previous, lastword]))
            word = self.db.random(key)
            if not word or word == self.stop_word:
                break
            out.append(word)
            previous = lastword
            lastword = word
        return ' '.join(out)

    def _split_message(self, msg):
        """Split message to better learn from it."""
        words = msg.split(' ')
        lastword = ""
        previous = ""
        msg += " " + self.stop_word

        for word in words:
            key = self._make_key(self.separator.join([previous, lastword]))
            self.db.add(key, word)
            yield [previous, lastword, word]
            previous = lastword
            lastword = word



from chattymarkov import ChattyMarkov
from chattymarkov.database import MemoryDatabase


class TestChattyMarkov:
    def test_init_memory(self):
        assert ChattyMarkov(MemoryDatabase()) is not None



#!/usr/bin/env python3
from os.path import abspath, dirname, join

from pip.req import parse_requirements
from setuptools import find_packages, setup

CWD = dirname(abspath(__file__))


def requires():
    """Parse the requirements.txt file and generate a requirements list."""
    # with open(join(CWD, 'requirements', 'base.txt'), 'r') as fp:
    install_reqs = parse_requirements(join(CWD, 'requirements', 'base.txt'),
                                      session=False)
    return [str(ir.req) for ir in install_reqs]


setup(
    author='Geoffrey ROYER',
    author_email='geoffrey.royer@gmail.com',
    name='chattymarkov',
    version="0.0.4",
    description='Generate random sentences through markov chains',
    url='https://github.com/Ge0/chattymarkov',
    license='MIT',
    include_package_data=True,
    packages=find_packages(),
    setup_requires=[
        'setuptools_scm'
    ],
    install_requires=requires()
)



"""
Redis database for chattimarkov.
"""
import redis

from .base import AbstractDatabase


class RedisDatabase(AbstractDatabase):
    def __init__(self, host="localhost", port=6739, db=0,
                 unix_socket_path=None):
        if unix_socket_path is not None:
            self.handle = redis.StrictRedis(
                unix_socket_path=unix_socket_path, db=db
            )
        else:
            self.handle = redis.StrictRedis(host=host, port=port, db=db)

    def add(self, key, element):
        return self.handle.sadd(key, element.encode()) > 0

    def random(self, key):
        element = self.handle.srandmember(key)
        if element is not None:
            return element.decode()

    def get(self, key):
        element = self.handle.get(key)
        if element is not None:
            return element.decode()

    def set(self, key, value):
        self.handle.set(key, value)
